@@ -5,13 +5,13 @@ import time
 import logging
 import warnings
 import percy
+import requests
 
 from selenium import webdriver
 from selenium.webdriver.support import expected_conditions as EC
 from selenium.webdriver.common.by import By
 from selenium.webdriver.support.wait import WebDriverWait
 from selenium.webdriver.common.keys import Keys
-from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
 from selenium.webdriver.common.action_chains import ActionChains
 
 from selenium.common.exceptions import (
@@ -20,7 +20,14 @@ from selenium.common.exceptions import (
     MoveTargetOutOfBoundsException,
 )
 
-from dash.testing.wait import text_to_equal, style_to_equal, contains_text, until
+from dash.testing.wait import (
+    text_to_equal,
+    style_to_equal,
+    class_to_equal,
+    contains_text,
+    contains_class,
+    until,
+)
 from dash.testing.dash_page import DashPageMixin
 from dash.testing.errors import DashAppLoadingError, BrowserError, TestingTimeoutError
 from dash.testing.consts import SELENIUM_GRID_DEFAULT
@@ -77,10 +84,10 @@ class Browser(DashPageMixin):
             )
             self.percy_runner.initialize_build()
 
-        logger.info("initialize browser with arguments")
-        logger.info("  headless => %s", self._headless)
-        logger.info("  download_path => %s", self._download_path)
-        logger.info("  percy asset root => %s", os.path.abspath(percy_assets_root))
+        logger.debug("initialize browser with arguments")
+        logger.debug("  headless => %s", self._headless)
+        logger.debug("  download_path => %s", self._download_path)
+        logger.debug("  percy asset root => %s", os.path.abspath(percy_assets_root))
 
     def __enter__(self):
         return self
@@ -103,20 +110,27 @@ class Browser(DashPageMixin):
         resource_path,
         hook_id,
         wait_for_callbacks=True,
+        convert_canvases=False,
         assert_check=True,
         stay_on_page=False,
+        widths=None,
     ):
         try:
             path = resource_path.lstrip("/")
             if path != resource_path:
                 logger.warning("we stripped the left '/' in resource_path")
-            self.driver.get("{}/{}".format(self.server_url.rstrip("/"), path))
+            self.driver.get(f"{self.server_url.rstrip('/')}/{path}")
 
             # wait for the hook_id to present and all callbacks get fired
             self.wait_for_element_by_id(hook_id)
-            self.percy_snapshot(path, wait_for_callbacks=wait_for_callbacks)
+            self.percy_snapshot(
+                path,
+                wait_for_callbacks=wait_for_callbacks,
+                convert_canvases=convert_canvases,
+                widths=widths,
+            )
             if assert_check:
-                assert not self.driver.find_elements_by_css_selector(
+                assert not self.find_elements(
                     "div.dash-debug-alert"
                 ), "devtools should not raise an error alert"
             if not stay_on_page:
@@ -125,14 +139,28 @@ class Browser(DashPageMixin):
             logger.exception("snapshot at resource %s error", path)
             raise e
 
-    def percy_snapshot(self, name="", wait_for_callbacks=False):
+    def percy_snapshot(
+        self, name="", wait_for_callbacks=False, convert_canvases=False, widths=None
+    ):
         """percy_snapshot - visual test api shortcut to `percy_runner.snapshot`.
-        It also combines the snapshot `name` with the Python version.
+        It also combines the snapshot `name` with the Python version,
+        args:
+        - name: combined with the python version to give the final snapshot name
+        - wait_for_callbacks: default False, whether to wait for Dash callbacks,
+            after an extra second to ensure that any relevant callbacks have
+            been initiated
+        - convert_canvases: default False, whether to convert all canvas elements
+            in the DOM into static images for percy to see. They will be restored
+            after the snapshot is complete.
+        - widths: a list of pixel widths for percy to render the page with. Note
+            that this does not change the browser in which the DOM is constructed,
+            so the width will only affect CSS, not JS-driven layout.
+            Defaults to [1280]
         """
-        snapshot_name = "{} - py{}.{}".format(
-            name, sys.version_info.major, sys.version_info.minor
-        )
-        logger.info("taking snapshot name => %s", snapshot_name)
+        if widths is None:
+            widths = [1280]
+
+        logger.info("taking snapshot name => %s", name)
         try:
             if wait_for_callbacks:
                 # the extra one second sleep adds safe margin in the context
@@ -148,7 +176,46 @@ class Browser(DashPageMixin):
                 self.redux_state_rqs,
             )
 
-        self.percy_runner.snapshot(name=snapshot_name)
+        if convert_canvases:
+            self.driver.execute_script(
+                """
+                const stash = window._canvasStash = [];
+                Array.from(document.querySelectorAll('canvas')).forEach(c => {
+                    const i = document.createElement('img');
+                    i.src = c.toDataURL();
+                    i.width = c.width;
+                    i.height = c.height;
+                    i.setAttribute('style', c.getAttribute('style'));
+                    i.className = c.className;
+                    i.setAttribute('data-canvasnum', stash.length);
+                    stash.push(c);
+                    c.parentElement.insertBefore(i, c);
+                    c.parentElement.removeChild(c);
+                });
+            """
+            )
+
+        try:
+            self.percy_runner.snapshot(name=name, widths=widths)
+        except requests.HTTPError as err:
+            # Ignore retries.
+            if err.request.status_code != 400:
+                raise err
+
+        if convert_canvases:
+            self.driver.execute_script(
+                """
+                const stash = window._canvasStash;
+                Array.from(
+                    document.querySelectorAll('img[data-canvasnum]')
+                ).forEach(i => {
+                    const c = stash[+i.getAttribute('data-canvasnum')];
+                    i.parentElement.insertBefore(c, i);
+                    i.parentElement.removeChild(i);
+                });
+                delete window._canvasStash;
+            """
+            )
 
     def take_snapshot(self, name):
         """Hook method to take snapshot when a selenium test fails. The
@@ -166,41 +233,57 @@ class Browser(DashPageMixin):
             except OSError:
                 logger.exception("cannot make artifacts")
 
-        self.driver.save_screenshot(
-            "{}/{}_{}.png".format(target, name, self.session_id)
-        )
+        self.driver.save_screenshot(f"{target}/{name}_{self.session_id}.png")
 
-    def find_element(self, selector):
-        """find_element returns the first found element by the css `selector`
-        shortcut to `driver.find_element_by_css_selector`."""
-        return self.driver.find_element_by_css_selector(selector)
-
-    def find_elements(self, selector):
-        """find_elements returns a list of all elements matching the css
-        `selector`.
-
-        shortcut to `driver.find_elements_by_css_selector`.
+    def find_element(self, selector, attribute="CSS_SELECTOR"):
+        """find_element returns the first found element by the attribute `selector`
+        shortcut to `driver.find_element(By.CSS_SELECTOR, ...)`.
+        args:
+        - attribute: the attribute type to search for, aligns with the Selenium
+            API's `By` class. default "CSS_SELECTOR"
+            valid values: "CSS_SELECTOR", "ID", "NAME", "TAG_NAME",
+            "CLASS_NAME", "LINK_TEXT", "PARTIAL_LINK_TEXT", "XPATH"
         """
-        return self.driver.find_elements_by_css_selector(selector)
+        return self.driver.find_element(getattr(By, attribute.upper()), selector)
+
+    def find_elements(self, selector, attribute="CSS_SELECTOR"):
+        """find_elements returns a list of all elements matching the attribute
+        `selector`. Shortcut to `driver.find_elements(By.CSS_SELECTOR, ...)`.
+        args:
+        - attribute: the attribute type to search for, aligns with the Selenium
+            API's `By` class. default "CSS_SELECTOR"
+            valid values: "CSS_SELECTOR", "ID", "NAME", "TAG_NAME",
+            "CLASS_NAME", "LINK_TEXT", "PARTIAL_LINK_TEXT", "XPATH"
+        """
+        return self.driver.find_elements(getattr(By, attribute.upper()), selector)
 
     def _get_element(self, elem_or_selector):
         if isinstance(elem_or_selector, str):
             return self.find_element(elem_or_selector)
         return elem_or_selector
 
-    def _wait_for(self, method, args, timeout, msg):
+    def _wait_for(self, method, timeout, msg):
         """Abstract generic pattern for explicit WebDriverWait."""
-        _wait = (
-            self._wd_wait if timeout is None else WebDriverWait(self.driver, timeout)
-        )
-        logger.debug(
-            "method, timeout, poll => %s %s %s",
-            method,
-            _wait._timeout,  # pylint: disable=protected-access
-            _wait._poll,  # pylint: disable=protected-access
-        )
+        try:
+            _wait = (
+                self._wd_wait
+                if timeout is None
+                else WebDriverWait(self.driver, timeout)
+            )
+            logger.debug(
+                "method, timeout, poll => %s %s %s",
+                method,
+                _wait._timeout,  # pylint: disable=protected-access
+                _wait._poll,  # pylint: disable=protected-access
+            )
 
-        return _wait.until(method(*args), msg)
+            return _wait.until(method)
+        except Exception as err:
+            if callable(msg):
+                message = msg(self.driver)
+            else:
+                message = msg
+            raise TimeoutException(message) from err
 
     def wait_for_element(self, selector, timeout=None):
         """wait_for_element is shortcut to `wait_for_element_by_css_selector`
@@ -212,12 +295,11 @@ class Browser(DashPageMixin):
         equals to the fixture's `wait_timeout` shortcut to `WebDriverWait` with
         `EC.presence_of_element_located`."""
         return self._wait_for(
-            EC.presence_of_element_located,
-            ((By.CSS_SELECTOR, selector),),
-            timeout,
-            "timeout {}s => waiting for selector {}".format(
-                timeout if timeout else self._wait_timeout, selector
+            EC.presence_of_element_located(
+                (By.CSS_SELECTOR, selector),
             ),
+            timeout,
+            f"timeout {timeout or self._wait_timeout}s => waiting for selector {selector}",
         )
 
     def wait_for_no_elements(self, selector, timeout=None):
@@ -227,10 +309,10 @@ class Browser(DashPageMixin):
             # if we use get_elements it waits a long time to see if they appear
             # so this one calls out directly to execute_script
             lambda: self.driver.execute_script(
-                "return document.querySelectorAll('{}').length".format(selector)
+                f"return document.querySelectorAll('{selector}').length"
             )
             == 0,
-            timeout if timeout else self._wait_timeout,
+            timeout or self._wait_timeout,
         )
 
     def wait_for_element_by_id(self, element_id, timeout=None):
@@ -238,12 +320,21 @@ class Browser(DashPageMixin):
         equals to the fixture's `wait_timeout` shortcut to `WebDriverWait` with
         `EC.presence_of_element_located`."""
         return self._wait_for(
-            EC.presence_of_element_located,
-            ((By.ID, element_id),),
-            timeout,
-            "timeout {}s => waiting for element id {}".format(
-                timeout if timeout else self._wait_timeout, element_id
+            EC.presence_of_element_located(
+                (By.ID, element_id),
             ),
+            timeout,
+            f"timeout {timeout or self._wait_timeout}s => waiting for element id {element_id}",
+        )
+
+    def wait_for_class_to_equal(self, selector, classname, timeout=None):
+        """Explicit wait until the element's class has expected `value` timeout
+        if not set, equals to the fixture's `wait_timeout` shortcut to
+        `WebDriverWait` with customized `class_to_equal` condition."""
+        return self._wait_for(
+            method=class_to_equal(selector, classname),
+            timeout=timeout,
+            msg=f"classname => {classname} not found within {timeout or self._wait_timeout}s",
         )
 
     def wait_for_style_to_equal(self, selector, style, val, timeout=None):
@@ -251,12 +342,9 @@ class Browser(DashPageMixin):
         if not set, equals to the fixture's `wait_timeout` shortcut to
         `WebDriverWait` with customized `style_to_equal` condition."""
         return self._wait_for(
-            method=style_to_equal,
-            args=(selector, style, val),
+            method=style_to_equal(selector, style, val),
             timeout=timeout,
-            msg="style val => {} {} not found within {}s".format(
-                style, val, timeout if timeout else self._wait_timeout
-            ),
+            msg=f"style val => {style} {val} not found within {timeout or self._wait_timeout}s",
         )
 
     def wait_for_text_to_equal(self, selector, text, timeout=None):
@@ -266,13 +354,25 @@ class Browser(DashPageMixin):
         shortcut to `WebDriverWait` with customized `text_to_equal`
         condition.
         """
+        method = text_to_equal(selector, text, timeout or self.wait_timeout)
+
         return self._wait_for(
-            method=text_to_equal,
-            args=(selector, text),
+            method=method,
             timeout=timeout,
-            msg="text -> {} not found within {}s".format(
-                text, timeout if timeout else self._wait_timeout
-            ),
+            msg=method.message,
+        )
+
+    def wait_for_contains_class(self, selector, classname, timeout=None):
+        """Explicit wait until the element's classes contains the expected `classname`.
+
+        timeout if not set, equals to the fixture's `wait_timeout`
+        shortcut to `WebDriverWait` with customized `contains_class`
+        condition.
+        """
+        return self._wait_for(
+            method=contains_class(selector, classname),
+            timeout=timeout,
+            msg=f"classname -> {classname} not found inside element within {timeout or self._wait_timeout}s",
         )
 
     def wait_for_contains_text(self, selector, text, timeout=None):
@@ -282,13 +382,11 @@ class Browser(DashPageMixin):
         shortcut to `WebDriverWait` with customized `contains_text`
         condition.
         """
+        method = contains_text(selector, text, timeout or self.wait_timeout)
         return self._wait_for(
-            method=contains_text,
-            args=(selector, text),
+            method=method,
             timeout=timeout,
-            msg="text -> {} not found inside element within {}s".format(
-                text, timeout if timeout else self._wait_timeout
-            ),
+            msg=method.message,
         )
 
     def wait_for_page(self, url=None, timeout=10):
@@ -302,35 +400,29 @@ class Browser(DashPageMixin):
             self.wait_for_element_by_css_selector(
                 self.dash_entry_locator, timeout=timeout
             )
-        except TimeoutException:
+        except TimeoutException as exc:
             logger.exception("dash server is not loaded within %s seconds", timeout)
-            logger.debug(self.get_logs())
+            logs = "\n".join((str(log) for log in self.get_logs()))
+            logger.debug(logs)
+            html = self.find_element("body").get_property("innerHTML")
             raise DashAppLoadingError(
                 "the expected Dash react entry point cannot be loaded"
-                " in browser\n HTML => {}\n Console Logs => {}\n".format(
-                    self.driver.find_element_by_tag_name("body").get_property(
-                        "innerHTML"
-                    ),
-                    "\n".join((str(log) for log in self.get_logs())),
-                )
-            )
+                f" in browser\n HTML => {html}\n Console Logs => {logs}\n"
+            ) from exc
 
         if self._pause:
-            try:
-                import pdb as pdb_
-            except ImportError:
-                import ipdb as pdb_
+            import pdb  # pylint: disable=import-outside-toplevel
 
-            pdb_.set_trace()
+            pdb.set_trace()  # pylint: disable=forgotten-debug-statement
 
     def select_dcc_dropdown(self, elem_or_selector, value=None, index=None):
         dropdown = self._get_element(elem_or_selector)
         dropdown.click()
 
-        menu = dropdown.find_element_by_css_selector("div.Select-menu-outer")
+        menu = dropdown.find_element(By.CSS_SELECTOR, "div.Select-menu-outer")
         logger.debug("the available options are %s", "|".join(menu.text.split("\n")))
 
-        options = menu.find_elements_by_css_selector("div.VirtualizedSelectOption")
+        options = menu.find_elements(By.CSS_SELECTOR, "div.VirtualizedSelectOption")
         if options:
             if isinstance(index, int):
                 options[index].click()
@@ -362,16 +454,11 @@ class Browser(DashPageMixin):
     def open_new_tab(self, url=None):
         """Open a new tab in browser url is not set, equals to `server_url`."""
         self.driver.execute_script(
-            'window.open("{}", "new window")'.format(
-                self.server_url if url is None else url
-            )
+            f'window.open("{url or self.server_url}", "new window")'
         )
 
     def get_webdriver(self):
-        try:
-            return getattr(self, "_get_{}".format(self._browser))()
-        except WebDriverException:
-            logger.exception("<<<Webdriver not initialized correctly>>>")
+        return getattr(self, f"_get_{self._browser}")()
 
     def _get_wd_options(self):
         options = (
@@ -388,8 +475,8 @@ class Browser(DashPageMixin):
     def _get_chrome(self):
         options = self._get_wd_options()
 
-        capabilities = DesiredCapabilities.CHROME
-        capabilities["loggingPrefs"] = {"browser": "SEVERE"}
+        options.set_capability("loggingPrefs", {"browser": "SEVERE"})
+        options.set_capability("goog:loggingPrefs", {"browser": "SEVERE"})
 
         if "DASH_TEST_CHROMEPATH" in os.environ:
             options.binary_location = os.environ["DASH_TEST_CHROMEPATH"]
@@ -404,15 +491,15 @@ class Browser(DashPageMixin):
                 "safebrowsing.disable_download_protection": True,
             },
         )
+        options.add_argument("--disable-dev-shm-usage")
+        options.add_argument("--no-sandbox")
+        options.add_argument("--disable-gpu")
+        options.add_argument("--remote-debugging-port=9222")
 
         chrome = (
-            webdriver.Remote(
-                command_executor=self._remote_url,
-                options=options,
-                desired_capabilities=capabilities,
-            )
+            webdriver.Remote(command_executor=self._remote_url, options=options)
             if self._remote
-            else webdriver.Chrome(options=options, desired_capabilities=capabilities)
+            else webdriver.Chrome(options=options)
         )
 
         # https://bugs.chromium.org/p/chromium/issues/detail?id=696481
@@ -435,15 +522,12 @@ class Browser(DashPageMixin):
     def _get_firefox(self):
         options = self._get_wd_options()
 
-        capabilities = DesiredCapabilities.FIREFOX
-        capabilities["loggingPrefs"] = {"browser": "SEVERE"}
-        capabilities["marionette"] = True
+        options.set_capability("loggingPrefs", {"browser": "SEVERE"})
+        options.set_capability("marionette", True)
 
-        # https://developer.mozilla.org/en-US/docs/Download_Manager_preferences
-        fp = webdriver.FirefoxProfile()
-        fp.set_preference("browser.download.dir", self.download_path)
-        fp.set_preference("browser.download.folderList", 2)
-        fp.set_preference(
+        options.set_preference("browser.download.dir", self.download_path)
+        options.set_preference("browser.download.folderList", 2)
+        options.set_preference(
             "browser.helperApps.neverAsk.saveToDisk",
             "application/octet-stream",  # this MIME is generic for binary
         )
@@ -451,22 +535,21 @@ class Browser(DashPageMixin):
             webdriver.Remote(
                 command_executor=self._remote_url,
                 options=options,
-                desired_capabilities=capabilities,
             )
             if self._remote
-            else webdriver.Firefox(
-                firefox_profile=fp, options=options, capabilities=capabilities
-            )
+            else webdriver.Firefox(options=options)
         )
 
     @staticmethod
     def _is_windows():
         return sys.platform == "win32"
 
-    def multiple_click(self, elem_or_selector, clicks):
+    def multiple_click(self, elem_or_selector, clicks, delay=None):
         """multiple_click click the element with number of `clicks`."""
         for _ in range(clicks):
             self._get_element(elem_or_selector).click()
+            if delay:
+                time.sleep(delay)
 
     def clear_input(self, elem_or_selector):
         """Simulate key press to clear the input."""
@@ -520,7 +603,7 @@ class Browser(DashPageMixin):
 
         Chrome only
         """
-        if self.driver.name.lower() == "chrome":
+        if self._browser == "chrome":
             return [
                 entry
                 for entry in self.driver.get_log("browser")
@@ -530,8 +613,8 @@ class Browser(DashPageMixin):
         return None
 
     def reset_log_timestamp(self):
-        """reset_log_timestamp only work with chrome webdrier."""
-        if self.driver.name.lower() == "chrome":
+        """reset_log_timestamp only work with chrome webdriver."""
+        if self._browser == "chrome":
             entries = self.driver.get_log("browser")
             if entries:
                 self._last_ts = entries[-1]["timestamp"]
@@ -562,3 +645,12 @@ class Browser(DashPageMixin):
     @property
     def download_path(self):
         return self._download_path
+
+    @property
+    def wait_timeout(self):
+        return self._wait_timeout
+
+    @wait_timeout.setter
+    def wait_timeout(self, value):
+        self._wait_timeout = value
+        self._wd_wait = WebDriverWait(self.driver, value)
